"""Initialization module for a data provider."""

from optitrader.market.investment_universe import InvestmentUniverse
from optitrader.market.market_data import MarketData

__all__ = [
    "MarketData",
    "InvestmentUniverse",
]
